//
//  BPlusTreeNode.cpp
//  B+Tree
//
//  Created by 葉朝鈞 on 2021/3/30.
//  Copyright © 2021 葉朝鈞. All rights reserved.
//

#include <iostream>
#include <algorithm>
#include <queue>
#include "BPlusTreeNode.hpp"
#include "BPlusTree.hpp"

using namespace std;

BPlusTreeNode::BPlusTreeNode(int a, bool leaf) {
    this->a = a;
    this->leaf = leaf;
    this->childs = new BPlusTreeNode*[a << 1] { NULL };
    this->keys = new int[(a << 1) - 1] { NULL };
    this->n = 0;
    this->prev = this->next = NULL;
}

int BPlusTreeNode::findKey(int k) {
    int i = 0;
    while (i < n && keys[i] < k)
        i++;
    return i;
}

void BPlusTreeNode::traverse() {
    queue<pair<int, BPlusTreeNode*>> q;
    int dep = 1, num = 0;
    q.push(pair<int, BPlusTreeNode*>(1, this));
    while (!q.empty()) {
        pair<int, BPlusTreeNode*> elem = q.front(); q.pop();
        int level = elem.first;
        BPlusTreeNode* x = elem.second;
        if (dep != level) {
            dep++;
            cout << endl;
        }
        for (int i = 0; i < x->n; i++)
            cout << x->keys[i] << " ";
        cout << "| ";
        if (!x->leaf)
            for (int i = 0; i <= x->n; i++)
                q.push(pair<int, BPlusTreeNode*>(level + 1, x->childs[i]));
        else
            num += x->n;
        
    }
    cout << endl << "[INFO] Recursived Traverse, size = " << num << endl;
    
    BPlusTreeNode* it;
    
    int num2 = 0;
    BPlusTreeNode* head = this;
    while (!head->leaf)
        head = head->childs[0];
    
    it = head;
    while (it != NULL) {
        num2 += it->n;
        for (int i = 0; i < it->n; i++)
            cout << it->keys[i] << " ";
        cout << "| ";
        it = it->next;
    }
    
    cout << endl << "[INFO] Linked Traverse, size = " << num2 << endl;
    
    int num3 = 0;
    BPlusTreeNode* tail = this;
    while (!tail->leaf)
        tail = tail->childs[tail->n];
    
    it = tail;
    while (it != NULL) {
        num3 += it->n;
        for (int i = 0; i < it->n; i++)
            cout << it->keys[i] << " ";
        cout << "| ";
        it = it->prev;
    }
    
    cout << endl << "[INFO] Back Linked Traverse, size = " << num3 << endl;
}

BPlusTreeNode* BPlusTreeNode::lookup(int k) {
    int i = findKey(k);
    if (i < n && keys[i] == k)
        return this;
    else if (leaf)
        return NULL;
    else
        return childs[i]->lookup(k);
}

int BPlusTreeNode::maxKey() {
    BPlusTreeNode* it = this;
    while (!it->leaf)
        it = it->childs[it->n];
    return it->keys[it->n - 1];
}

int BPlusTreeNode::minKey() {
    BPlusTreeNode* it = this;
    while (!it->leaf)
        it = it->childs[0];
    return it->keys[0];
}

void BPlusTreeNode::insert(int k) {
    if (leaf) {
        int i = n;
        while (i > 0 && keys[i - 1] > k) {
            keys[i] = keys[i - 1];
            i--;
        }
        keys[i] = k;
        n++;
    } else {
        int p = findKey(k);
        if (childs[p]->n == (a << 1) - 1) {
            splitChild(p, childs[p]);
            p += (k < keys[p] ? 0 : 1);
        }
        childs[p]->insert(k);
    }
}

void BPlusTreeNode::splitChild(int p, BPlusTreeNode *y) {
    BPlusTreeNode* z = new BPlusTreeNode(a, y->leaf);
    if (y->leaf) {
        
        y->n = a; z->n = a - 1;
        
        for (int i = 0; i < z->n; i++)
            z->keys[i] = y->keys[a + i];
        for (int i = 0; i < z->n + 1; i++)
            z->childs[i] = y->childs[a + i];

        for (int i = n; i > p; i--)
            keys[i] = keys[i - 1];
        for (int i = n + 1; i > p; i--)
            childs[i] = childs[i - 1];
        
        keys[p] = y->keys[y->n - 1]; 
        
        z->next = y->next;
        if (y->next != NULL)
            y->next->prev = z;
        y->next = z;
        z->prev = y;
        
    } else {
        
        y->n = z->n = a - 1;

        for (int i = 0; i < z->n; i++)
            z->keys[i] = y->keys[a + i];
        for (int i = 0; i < z->n + 1; i++)
            z->childs[i] = y->childs[a + i];
        
        for (int i = n; i > p; i--)
            keys[i] = keys[i - 1];
        for (int i = n + 1; i > p; i--)
            childs[i] = childs[i - 1];
        
        keys[p] = y->maxKey();
    }
    childs[p] = y;
    childs[p + 1] = z;
    n++;
}

void BPlusTreeNode::remove(int k) {
    int p = findKey(k);
    
    if (leaf) {
        if (p < n && keys[p] == k) {
            for (int i = p; i < n - 1; i++)
                keys[i] = keys[i + 1];
            n--;
        } else {
            throw exception();
        }
    } else {
        childs[p]->remove(k);
    }

    // 索引关键字被删，取左孩子的最大值替换（索引关键字规则）
    if (!leaf && keys[p] == k)
        keys[p] = childs[p]->maxKey();
    
    if (!leaf && childs[p]->n < a - 1) {
        // 1. 借用
        // 右节点：借用后索引节点要变，更新为旧的右侧第一个
        // 左节点：索引也要变，更新为左节点最后一个（需要计算）
        
        // 2. 合并
        // 叶子节点直接借用，更新左孩子的索引关键字到右边的借用值
        // 非叶子节点等价于B树，拉下父节点的关键字填补
        BPlusTreeNode* l = p > 0 ? childs[p - 1] : NULL;
        BPlusTreeNode* r = p < n ? childs[p + 1] : NULL;
        
        if (r != NULL && r->n >= a) {
            borrowFromRight(p);
        } else if (l != NULL && l->n >= a) {
            borrowFromLeft(p);
        } else {
            // 叶子节点的可合并数是非叶子节点多一个，因为不需要下拉父索引关键字
            int w = (l == NULL ? r->leaf : l->leaf) ? 0 : 1;
            
            if (r != NULL && r->n <= a - w)
                mergeToLeft(p);
            else if (l != NULL && l->n <= a - w)
                mergeToLeft(p - 1);
            else {
                // should be impossible
                throw exception();
            }
        }
    }
}

void BPlusTreeNode::mergeToLeft(int p) {
    BPlusTreeNode* l = childs[p];
    BPlusTreeNode* r = childs[p + 1];
    
    if (l->leaf) {
        
        for (int i = 0; i < r->n; i++)
            l->keys[l->n + i] = r->keys[i];
        
        l->n += r->n;
        
        l->next = r->next;
    
    } else {
        
        l->keys[l->n] = keys[p];
        
        for (int i = 0; i < r->n; i++)
            l->keys[l->n + 1 + i] = r->keys[i];
        for (int i = 0; i < r->n + 1; i++)
            l->childs[l->n + 1 + i] = r->childs[i];
        
        l->n += r->n + 1;

    }

    for (int i = p; i < n - 1; i++)
        keys[i] = keys[i + 1];
    for (int i = p + 1; i < n; i++)
        childs[i] = childs[i + 1];
    
    n--;
    delete r;
}

void BPlusTreeNode::borrowFromLeft(int p) {
    BPlusTreeNode* l = childs[p - 1];
    BPlusTreeNode* r = childs[p];
    
    if (l->leaf) {
        
        for (int i = r->n; i > 0; i--)
            r->keys[i] = r->keys[i - 1];
        
        r->keys[0] = l->keys[l->n - 1];
        
        l->n--;
        r->n++;
        
        keys[p - 1] = l->keys[l->n - 1];
        
    } else {
        
        for (int i = r->n; i > 0; i--)
            r->keys[i] = r->keys[i - 1];
        for (int i = r->n + 1; i > 0; i--)
            r->childs[i] = r->childs[i - 1];
        
        r->childs[0] = l->childs[l->n];
        r->keys[0] = r->childs[0]->maxKey();
        
        l->n--;
        r->n++;
        
        keys[p - 1] = l->maxKey();
        
    }
}

void BPlusTreeNode::borrowFromRight(int p) {
    BPlusTreeNode* l = childs[p];
    BPlusTreeNode* r = childs[p + 1];
    
    if (l->leaf) {
        
        l->keys[l->n] = r->keys[0];
        keys[p] = r->keys[0];
        
        for (int i = 0; i < r->n - 1; i++)
            r->keys[i] = r->keys[i + 1];
        
        l->n++;
        r->n--;
        
    } else {
        
        l->keys[l->n] = l->childs[l->n]->maxKey(); // 左侧其实我们增加了一个关键字
        l->childs[l->n + 1] = r->childs[0];
        
        keys[p] = r->keys[0];
        
        for (int i = 0; i < r->n - 1; i++)
            r->keys[i] = r->keys[i + 1];
        for (int i = 0; i < r->n; i++)
            r->childs[i] = r->childs[i + 1];

        l->n++;
        r->n--;
        
    }
}
