//
//  BTreeNode.cpp
//  BTree
//
//  Created by 葉朝鈞 on 2019/2/16.
//  Copyright © 2019 TalentEndless. All rights reserved.
//

#include <iostream>
#include "BTreeNode.hpp"

using namespace std;

BTreeNode::BTreeNode(int a, bool leaf) {
    this->a = a;
    this->leaf = leaf;
    this->childs = new BTreeNode*[a << 1] { NULL };
    this->keys = new int[(a << 1) - 1] { NULL };
    this->n = 0;
}

int BTreeNode::findKey(int k) {
    int i = 0;
    while (i < n && keys[i] < k)
        i++;
    return i;
}

int BTreeNode::traverse() {
    int num = n;
    for (int i = 0; i <= n; i++) {
        if (!leaf)
            num += childs[i]->traverse();
        if (i < n)
            cout << keys[i] << ", ";
    }
    return num;
}

BTreeNode* BTreeNode::lookup(int k) {
    int i = findKey(k);
    if (i < n && keys[i] == k)
        return this;
    else if (leaf)
        return NULL;
    else
        return childs[i]->lookup(k);
}

int BTreeNode::maxKey() {
    BTreeNode* it = this;
    while (!it->leaf)
        it = it->childs[it->n];
    return it->keys[it->n - 1];
}

int BTreeNode::minKey() {
    BTreeNode* it = this;
    while (!it->leaf)
        it = it->childs[0];
    return it->keys[0];
}

void BTreeNode::insert(int k) {
    if (leaf) {
        int i = n;
        while (i > 0 && keys[i - 1] > k) {
            keys[i] = keys[i - 1];
            i--;
        }
        keys[i] = k;
        n++;
    } else {
        int p = findKey(k);
        // 由于分裂需要往父节点添加，为避免父节点可能已满导致的不断回溯，
        // 因而在递归之前进行分裂，但此做法可能效率低下
        if (childs[p]->n == (a << 1) - 1) {
            splitChild(p, childs[p]);
            p += (k < keys[p] ? 0 : 1);
        }
        childs[p]->insert(k);
    }
}

void BTreeNode::splitChild(int p, BTreeNode *y) {
    BTreeNode* z = new BTreeNode(y->a, y->leaf);

    z->n = y->n = a - 1;
    
    // n KEYS
    for (int i = 0; i < z->n; i++)
        z->keys[i] = y->keys[a + i];
    // n KEYS means n + 1 CHILDS
    for (int i = 0; i < z->n + 1; i++)
        z->childs[i] = y->childs[a + i];
    
    for (int i = n; i > p; i--)
        keys[i] = keys[i - 1];
    for (int i = n + 1; i > p; i--)
        childs[i] = childs[i - 1];
    
    keys[p] = y->keys[a - 1];
    childs[p] = y;
    childs[p + 1] = z;
    
    n++;
}

void BTreeNode::remove(int k) {
    int p = findKey(k);
    
    if (p < n && keys[p] == k) {
        if (leaf)
            removeFromLeaf(p);
        else
            removeFromNonLeaf(p);
    } else {
        if (leaf)
            throw exception();
        else
            childs[p]->remove(k);
    }

    // 删除后，其子节点度数小于最小阶，则需要借节点或者合并
    // 当前需要非叶子节点，如果叶子，则返回上一层处理
    if (!leaf && childs[p]->n < a - 1) {
        if (p < n && childs[p + 1] != NULL && childs[p + 1]->n >= a) {
            borrowFromRight(p);
        } else if (p > 0 && childs[p - 1] != NULL && childs[p - 1] ->n >= a) {
            borrowFromLeft(p);
        } else {
            if (p < n && childs[p + 1] != NULL && childs[p + 1] ->n <= a - 1)
                mergeToLeft(p);
            else if (p > 0 && childs[p - 1] != NULL && childs[p - 1] ->n <= a - 1)
                mergeToLeft(p - 1);
            else
                throw exception();
        }
    }
}

void BTreeNode::removeFromLeaf(int p) {
    for (int i = p; i < n - 1; i++)
        keys[i] = keys[i + 1];
    n--;
}

void BTreeNode::removeFromNonLeaf(int p) {
    BTreeNode* l = childs[p];
    BTreeNode* r = childs[p + 1];
    if (l->n >= a) {
        int key = l->maxKey();
        keys[p] = key;
        l->remove(key);
    } else if (r->n >= a) {
        int key = r->minKey();
        keys[p] = key;
        r->remove(key);
    } else {
        int k = keys[p];
        mergeToLeft(p);
        remove(k);
    }
}

void BTreeNode::mergeToLeft(int p) {
    BTreeNode* l = childs[p];
    BTreeNode* r = childs[p + 1];
    
    l->keys[l->n] = keys[p];
    
    for (int i = 0; i < r->n; i++)
        l->keys[l->n + 1 + i] = r->keys[i];
    
    for (int i = 0; i < r->n + 1; i++)
        l->childs[l->n + 1 + i] = r->childs[i];
    
    l->n += r->n + 1;
    
    for (int i = p; i < n - 1; i++)
        keys[i] = keys[i + 1];
    
    for (int i = p + 1; i < n; i++)
        childs[i] = childs[i + 1];
    
    n--;
    
    delete r;
}

// 左借，相当于右旋
void BTreeNode::borrowFromLeft(int p) {
    BTreeNode* l = childs[p - 1];
    BTreeNode* r = childs[p];
    int k = l->keys[l->n - 1];
    BTreeNode* rtree = l->childs[l->n];

    // 左借后，左边n--，右边需要腾空借入的位置
    for (int i = r->n; i > 0 ; i--)
        r->keys[i] = r->keys[i - 1];
    for (int i = childs[p]->n + 1; i > 0 ; i--)
        r->childs[i] = r->childs[i - 1];

    l->n--;

    r->keys[0] = keys[p - 1]; // 左替换根相邻的数据
    keys[p - 1] = k;
    r->childs[0] = rtree; // 左子树挂上右子树

    r->n++;
}

// 右借，相当于左旋
void BTreeNode::borrowFromRight(int p) {
    BTreeNode* l = childs[p];
    BTreeNode* r = childs[p + 1];
    int k = r->keys[0];
    BTreeNode* ltree = r->childs[0];

    // 右借后，左边n++，右边需要填补借出的位置
    for (int i = 0; i < r->n - 1; i++)
        r->keys[i] = r->keys[i + 1];
    for (int i = 0; i < childs[p + 1]->n; i++)
        r->childs[i] = r->childs[i + 1];

    r->n--;

    l->keys[l->n] = keys[p]; // 替换根相邻的数据
    keys[p] = k;
    l->childs[l->n + 1] = ltree; // 左子树挂在右子树上

    l->n++;
}
